package Nagios::Check::DNS::check_tlsa_record;

use strict;
use warnings;
use feature qw(say switch);
use if $ENV{DEBUG} => 'Smart::Comments';
use Carp;
use Data::Dumper;

#use if $^V >= v5.0.20 => (experimental => gw(smartmatch));
use experimental qw(smartmatch);
use Fcntl qw(F_GETFD F_SETFD FD_CLOEXEC);

our $VERSION = '0.1';

my $dane_pattern =
qr'(?i)^(?<record>(?<tlsa_usage>\d+)\s+(?<tlsa_selector>\d+)
\s+(?<tlsa_match_type>\d+)\s+(?<tlsa_hash>[0-9a-f ]+))$'xs;

# Alternativly my $tmpfile = File::Temp->new();
#              unlink($tmpfile);
# if we need more # control over the directory, the name, and such.
# The sub _tmpfile_ has the sole purpose to use the $tmpfile in a
# *global* context. This prevents our garbage collector from closing the
# file!
open(my $tmpfile, '+>', undef)
  or die "Can't open anonymous file: $!\n";
fcntl($tmpfile, F_SETFD, fcntl($tmpfile, F_GETFD, 0) & ~FD_CLOEXEC)
  or die "clear FD_CLOEXEC on $tmpfile: $!\n";
sub __tmpfile { $tmpfile }
my $fdname = '/dev/fd/' . fileno $tmpfile;

sub main {
    my $domain   = shift;
    my $port     = shift // 443;
    my $protocol = shift // 'tcp';
    my $expiry   = shift;

    my @validate = validate_tlsa($domain, $port, $protocol, $expiry);


    return join("\n", @validate);

}

sub get_tlsa_from_dns {
    my $domain     = shift;
    my $port       = shift // croak 'Need a port number';
    my $protocol   = shift // 'tcp';
    my $query      = "dig tlsa _$port._$protocol.$domain +short";
    my @dns_return = qx($query);
    my @tlsa_line;

    foreach my $item (@dns_return) {
        if ($item !~ /^[_a-z]+[a-z0-9]+/i) {
            chomp $item;
            push @tlsa_line, $item;
        }
    }

    return @tlsa_line;

}

sub get_cert {
    my $domain = shift;
    my $port   = shift;
    my $cmd;
    my $cert;

    if ($port == 25) {
        $cmd = "openssl s_client -starttls smtp -connect $domain:$port";
    }
    else {
        $cmd = "openssl s_client -connect $domain:$port";
    }
    my $same = "< /dev/null 2>/dev/null | openssl x509 -out $fdname 2>&1";
    $cmd .= $same;

    die sprintf "[%s] returned exit:%d signal:%d\n", $cmd, $? >> 8, $? & 0xff
      if $?;

    return qx($cmd);
}

sub get_hash_from_cert {
    my $cert          = shift;
    my $hashit        = shift // 'sha256';
    my $tlsa_selector = shift // 1;
    my $gentlsa;

    $gentlsa = <<_;
openssl x509  -in $fdname -pubkey |
openssl rsa -pubin -inform PEM -outform DER 2>/dev/null |
openssl $hashit
_

    if ($tlsa_selector == 0) {
        $gentlsa = "openssl x509 -in $fdname -outform DER | openssl $hashit";
    }

    my $tlsa_record = qx($gentlsa) or die "nothing found!\n";
    $tlsa_record =~ s/^.*= (.*$)/$1/gi;
    $tlsa_record = uc($tlsa_record);
    chomp $tlsa_record;

    return $tlsa_record;
}

sub check_expiry {
    my $check_expiry = "openssl x509 -in $fdname -noout -startdate -enddate";
    my $expiry       = qx($check_expiry);
    $expiry =~ tr/\n/ /d;
    chomp $expiry;

    return $expiry;
}

sub get_tlsa_match_type {
    my $dig_return = shift;
    my $hashit;
    my $tlsa_match_type;

    if ($dig_return =~ /$dane_pattern/) {
        $tlsa_match_type = $+{tlsa_match_type};
    }

    if ($tlsa_match_type >= 3) {
        return "Not valid: $tlsa_match_type";
    }

    for ($tlsa_match_type) {
        when ('0') { die 'certs will be compared directly' }
        when ('1') { $hashit = 'sha256' }
        when ('2') { $hashit = 'sha512' }
        default { $hashit = 'sha256' }
    }
    return $hashit;
}

sub get_tlsa_dns_record {
    my $dns_return = shift;
    my $dns_tlsa;

    if ($dns_return =~ /$dane_pattern/i) {
        $dns_tlsa = $+{tlsa_hash};
        $dns_tlsa =~ s/(\S*)\s+(\S*)$/$1$2/;
    }
    return $dns_tlsa;
}

sub get_tlsa_usage {
  #
  #      @TODO: check certificate trust chain
  #      get_ca_cert()
  #      need to get ca cert for verification if tlsa usage < 2
  #   0: CA Constraints:
  #         Zertifikat muss von der angegebenen CA stammen.
  #         Der Hash wird aus dem Public Certificate der CA generiert.
  #         x509 Trust chain muss gültig sein.
  #   1: Certificate Constraints:
  #         Nur das angegebene Zertifikat darf mit der Domain eingesetzt werden.
  #         Hash wird aus diesem Zertifikat generiert.
  #         x509 Trust chain muss gültig sein.
  #   2: Trust anchor assertion:
  #         Das Zertifikat muss von der angegebenen CA stammen.
  #         Hash aus Public Cert der CA.
  #         Keine Trust chain-Überprüfung.
  #   3: Domain-Issued certificates:
  #         Nur das angegebene Zertifikat darf mit der Domain eingesetzt werden.
  #         Hash aus dem eigenen Public Cert.
  #         Keine Trust chain-Überprüfung.
  #
  # https://tools.ietf.org/html/rfc6698#section-2.1.1
  #
    my $dns_return = shift;
    my $tlsa_usage;

    if ($dns_return =~ /$dane_pattern/i) {
        $tlsa_usage = $+{tlsa_usage};
    }
    return $tlsa_usage;
}

sub get_tlsa_selector {
    #
    #    0: Full certificate: the Certificate binary structure as defined
    #          in [RFC5280]
    #    1: SubjectPublicKeyInfo: DER-encoded binary structure as defined
    #          in [RFC5280]
    #         Vorteil: Wenn immer derselbe Private Key für die Generierung von
    #         Zertifikaten genutzt wird, muss der TLSA-Record nicht mit jedem
    #         Zertifikatswechsel erneuert werden.
    #
    my $dns_return = shift;
    my $tlsa_selector;

    if ($dns_return =~ /$dane_pattern/i) {
        $tlsa_selector = $+{tlsa_selector};
    }

    return $tlsa_selector;

}

sub validate_tlsa {
    my $domain          = shift;
    my $port            = shift;
    my $protocol        = shift;
    my $expiry          = shift;
    my @dns_return      = get_tlsa_from_dns($domain, $port, $protocol);
    my $fail_selector   = 0;
    my $fail_usage      = 0;
    my $fail_match_type = 0;
    my @dns_tlsa;
    my @tlsa_selector;
    my @tlsa_usage;
    my @tlsa_match_type;
    my @return;

    if (!@dns_return) {
        return "UNKNOWN: No DANE to check for $domain:$port";
    }

    my $cert = get_cert($domain, $port);
    if ($cert =~ /.*unable to load certificate.*/) {
        return "WARNING: No SSL-Certificate available for $domain:$port";
    }


    foreach my $item (@dns_return) {

        my %domain = (
            'tlsa_dns'        => get_tlsa_dns_record($item),
            'tlsa_selector'   => get_tlsa_selector($item),
            'tlsa_usage'      => get_tlsa_usage($item),
            'tlsa_match_type' => get_tlsa_match_type($item),
        );

        my $tlsa_usage      = $domain{'tlsa_usage'};
        my $tlsa_selector   = $domain{'tlsa_selector'};
        my $tlsa_match_type = $domain{'tlsa_match_type'};
        my $dns_tlsa_hash   = $domain{'tlsa_dns'};
        my $cert_tlsa_hash;

        if ($tlsa_selector < 0 or $tlsa_selector > 1) {
            push @return, "CRITICAL: TLSA Selector \'$tlsa_selector\' for "
            . "$domain:$port is not valid";
            $fail_selector = 1;
        }

        if ($tlsa_usage < 0 or $tlsa_usage > 3) {
            push @return, "CRITICAL: TLSA Usage \'$tlsa_usage\' for "
            . "$domain:$port is not valid";
            $fail_usage = 1;
        }

        if ($tlsa_match_type =~ /not.*(?<mt>\d+)/i) {
            push @return, "CRITICAL: TLSA Match Type  \'$+{mt}\' for "
            . "$domain:$port is not valid";
            $fail_match_type = 1;
        }

        if ($fail_match_type != 1) {
            $cert_tlsa_hash  = get_hash_from_cert($cert, $tlsa_match_type);

            if ($fail_selector != 1) {
                $cert_tlsa_hash =
                  get_hash_from_cert($cert, $tlsa_match_type, $tlsa_selector);
            }

            #chomp $cert_tlsa_hash;
        }

        if ($fail_usage != 1 and $fail_selector != 1 and $fail_match_type != 1)
        {
            if ($dns_tlsa_hash ne $cert_tlsa_hash) {
                push @return,
                  "CRITICAL: DANE for $domain:$port is not valid";
            }
            else {
                push @return, "OK: DANE for $domain:$port is valid";
            }
        }

    }

    if ($expiry) {
      my $ex  = check_expiry();
      #push @return, "Cert $domain:$port valid: $ex";
      push @return, $ex;
    }

    return @return;
}

# vim: ft=perl ts=2 sw=2 foldmethod=indent
1;
